=pod

---+ package TWiki::Store::Mercurial

Wrapper around the Mercurial(hg) commands required by TWiki.
There is one of these object for each file stored under Mercurial.

=cut

package TWiki::Store::Mercurial;

use TWiki;
use TWiki::Time;
use TWiki::Sandbox;

use strict;
use Assert;

sub new {
    my( $class, $session, $web, $topic, $attachment ) = @_;
    #ASSERT($session->isa( 'TWiki')) if DEBUG;
    my $this = bless( {}, $class );
    $this->{session} = $session;

    $this->{web} = $web;

    $this->{revs} = [];

    if( $topic ) {
        $this->{topic} = $topic;

        if( $attachment ) {

            $this->{attachment} = $attachment;

            $this->{file} = $TWiki::cfg{PubDir}.'/'.$web.'/'.
              $this->{topic}.'/'.$attachment;

        } else {
            $this->{file} = $TWiki::cfg{DataDir}.'/'.$web.'/'.
              $topic.'.txt';
        }
    }

    return $this;
}

=pod

---++ ObjectMethod storedDataExists() -> $boolean
Establishes if there is stored data associated with this handler.

=cut

sub storedDataExists {
    my $this = shift;
    return -e $this->{file};
}

=pod

---++ ObjectMethod getLatestRevision() -> $text
Get the text of the most recent revision

=cut

sub getLatestRevision {
    my $this = shift;
    return $this->_readFile( $this->{file} );
}

=pod

---++ sub _readFile

=cut

sub _readFile {
    my( $this, $name ) = @_;
    my $data;
    if( open( IN_FILE, '<', $name )) {
        binmode( IN_FILE );
        local $/ = undef;
        $data = <IN_FILE>;
        close( IN_FILE );
    }
    $data ||= '';
    return $data;
}

=pod

---++ ObjectMethod getAttachmentList($web, $topic)

returns {} of filename => { key => value, key2 => value } for any given web, topic
Ignores files starting with _ or ending with .orig

=cut

sub getAttachmentList {
	my( $this, $web, $topic ) = @_;
	my $dir = dirForTopicAttachments($web, $topic);
		
    opendir DIR, $dir || return '';
    my %attachmentList = ();
    my @files = sort grep { m/^[^\.*_]/ } readdir( DIR );
    @files = grep { !/.*.orig/ } @files;
    foreach my $attachment ( @files ) {
    	my @stat = stat ($dir."/".$attachment);
        $attachmentList{$attachment} = _constructAttributesForAutoAttached($attachment, \@stat);
    }
    closedir( DIR );
    return %attachmentList;
}

sub dirForTopicAttachments {
    my ($web, $topic ) = @_;
    return $TWiki::cfg{PubDir}.'/'.$web.'/'.$topic;
}

=pod

sub _constructAttributesForAutoAttached
as long as stat is defined, return an emulated set of attributes for that attachment.

=cut

sub _constructAttributesForAutoAttached {
    my ($file, $stat) = @_;

    my %pairs = (
        name    => $file,
        version => '',
        path    => $file,
        size    => $stat->[7],
        date    => $stat->[9], 
        user    => 'UnknownUser', 
        comment => '',
        attr    => '',
        autoattached => '1'
       );

    if ($#$stat > 0) {
        return \%pairs;
    } else {
        return undef;
    }
}

=pod

---++ ObjectMethod getTopicNames() -> @topics

Get list of all topics in a web
   * =$web= - Web name, required, e.g. ='Sandbox'=
Return a topic list, e.g. =( 'WebChanges',  'WebHome', 'WebIndex', 'WebNotify' )=

=cut

sub getTopicNames {
    my $this = shift;

    opendir DIR, $TWiki::cfg{DataDir}.'/'.$this->{web};
    # the name filter is used to ensure we don't return filenames
    # that contain illegal characters as topic names.
    my @topicList =
      sort
        map { TWiki::Sandbox::untaintUnchecked( $_ ) }
          grep { !/$TWiki::cfg{NameFilter}/ && s/\.txt$// }
            readdir( DIR );
    closedir( DIR );
    return @topicList;
}

=pod

---++ ObjectMethod getWebNames() -> @webs
Gets a list of names of subwebs in the current web

=cut

sub getWebNames {
    my $this = shift;
    my $dir = $TWiki::cfg{DataDir}.'/'.$this->{web};
    if( opendir( DIR, $dir ) ) {
        my @tmpList =
          sort
            map { TWiki::Sandbox::untaintUnchecked( $_ ) }
              grep { !/\./ &&
                     !/$TWiki::cfg{NameFilter}/ &&
                     -d $dir.'/'.$_
                   }
                readdir( DIR );
        closedir( DIR );
        return @tmpList;
    }
    return ();
}

=pod

---++ ObjectMethod finish
Complete processing after the client's HTTP request has been responded
to.
   1 breaking circular references to allow garbage collection in persistent
     environments

=cut

sub finish {
    my $this = shift;
}

sub getRevision {
    my( $this, $version ) = @_;

#If the version number is greater than the amount of entries in the array from mercurial then change version to entries
    if ( !@{$this->{revs}} ) {
	$this->_getVersionArray();
    }
    if ( $version > @{$this->{revs}} ) {
        $version = @{$this->{revs}};
    }

    my $hgversion = $this->_getHgRev( $version );
    
    my $file = $this->{file};
    my ($text, $status) = $this->{session}->{sandbox}->sysCommand(
        $TWiki::cfg{Hg}{coCmd},
        REVISION => $hgversion,
        FILENAME => $file );
 
    return $text;
}

sub _getHgRev {
    my ($this,$version) = @_;
    return if $version < 1;
    $this->_getVersionArray();
    return ${$this->{revs}}[$version-1];
}

sub _getVersionArray {
    my ( $this ) = @_;
    
    if ( @{$this->{revs}} ) {
	return;
    }

    my ( $hgOut, $exit ) = $this->{session}->{sandbox}->sysCommand
       ( $TWiki::cfg{Hg}{numRevisions}, 
         FILENAME => $this->{file} );

    if( ! $exit ) {
    	if( $hgOut ) {
	    $hgOut =~ s/\'//g; 
	    my @revs = reverse(split (/,/,$hgOut));	
	    my $prevVersion = 'none';
	    my $rev;
	    foreach $rev (@revs){
		if ( $rev =~ m/^(.*):(.*)$/ ) {
	            #Make it compatible with old comments
		    if(( $prevVersion eq $2 ) && ( $2 =~ m/Version/ )){
		        pop(@{$this->{revs}});
		    } 
		    $prevVersion = $2;
		} else {
		    $prevVersion = 'none';
		}
            push(@{$this->{revs}}, $1);
	    }
	}
    }
}

sub numRevisions {
    my( $this ) = @_;
    $this->_getVersionArray();
    return ( @{$this->{revs}} );
}

sub getRevisionInfo {
    my( $this, $version ) = @_;
	my $hgversion = $this->_getHgRev( $version );
        my( $hgOut, $exit ) = $this->{session}->{sandbox}->sysCommand(
            $TWiki::cfg{Hg}{infoCmd},
            REVISION => $hgversion, 
            FILENAME => $this->{file} );
        if( ! $exit ) {
            if( $hgOut =~ /date:(.*)user:(.*)summary:(.*)$/ ) {
                my $user = $2;
                my $comment = $3;
                my $date = $1;
                my $rev = $version; #$hgversion;
            
                return( $rev, $date, $user, $comment );
            }
        }
}

sub revisionDiff {
    my( $this, $rev1, $rev2, $contextLines ) = @_;
    my $tmp = '';
    my $exit;
    if ( $rev1 eq '1' && $rev2 eq '1' ) {
        my $text = $this->getRevision(1);
        $tmp = "1a1\n";
        foreach( split( /\r?\n/, $text ) ) {
            $tmp = "$tmp> $_\n";
        }
    } else {
	$rev1 =  $this->_getHgRev( $rev1 );
        $rev2 =  $this->_getHgRev( $rev2 );
        $contextLines = 3 unless defined($contextLines);
        ( $tmp, $exit ) = $this->{session}->{sandbox}->sysCommand(
            $TWiki::cfg{Hg}{diffCmd},
            REVISION1 => $rev1,
            REVISION2 => $rev2,
            FILENAME => $this->{file},
            CONTEXT => $contextLines );
    }
    return parseRevisionDiff( $tmp );
#	return $tmp;
}

=pod

---++ StaticMethod parseRevisionDiff( $text ) -> \@diffArray

| Description: | parse the text into an array of diff cells |
| #Description: | unlike Algorithm::Diff I concatinate lines of the same diffType that are sqential (this might be something that should be left up to the renderer) |
| Parameter: =$text= | currently unified or rcsdiff format |
| Return: =\@diffArray= | reference to an array of [ diffType, $right, $left ] |

=cut

sub parseRevisionDiff {
    my( $text ) = @_;

    my ( @diffArray ) = ();

    $text =~ s/\r//go;  # cut CR

    my $lineNumber=1;
    foreach( split( /\r?\n/, $text ) ) {
	    if ( $lineNumber > 3 ) {   #skip the first 3 lines (filenames)
		    if ( /@@ [-+]([0-9]+)([,0-9]+)? [-+]([0-9]+)(,[0-9]+)? @@/ ) {
#line number
			    push @diffArray, ['l', $1, $3];
		    } elsif( /^\-(.*)$/ ) {
			    push @diffArray, ['-', $1, ''];
		    } elsif( /^\+(.*)$/ ) {
			    push @diffArray, ['+', '', $1];
		    } elsif( s/^ (.*)$/$1/go ) {
			    push @diffArray, ['u', $_, $_];
		    }
	    }
	    $lineNumber++;
    }
    return \@diffArray;
}

=pod

---++ ObjectMethod getLatestRevisionTime() -> $text
Get the time of the most recent revision

=cut

sub getLatestRevisionTime {
    my @e = stat( shift->{file} );
    return $e[9] || 0;
}

=pod

---++ ObjectMethod getTimestamp() -> $integer

Get the timestamp of the file
Returns 0 if no file, otherwise epoch seconds

=cut

sub getTimestamp {
    my( $this ) = @_;
    my $date = 0;
    if( -e $this->{file} ) {
        # SMELL: Why big number if fail?
        $date = (stat $this->{file})[9] || 600000000;
    }
    return $date;
}

=pod

---++ ObjectMethod setLease( $lease )
   * =$lease= reference to lease hash, or undef if the existing lease is to be cleared.

Set an lease on the topic.

=cut

sub setLease {
    my( $this, $lease ) = @_;

    my $filename = $this->_controlFileName('lease');
    if( $lease ) {
        $this->_saveFile( $filename, join( "\n", %$lease ) );
    } elsif( -e $filename ) {
        unlink $filename ||
          throw Error::Simple( 'HG: failed to delete '.$filename.': '.$! );
    }
}


=pod

---++ ObjectMethod getLease() -> $lease

Get the current lease on the topic.

=cut

sub getLease {
    my( $this ) = @_;

    my $filename = $this->_controlFileName('lease');
    if ( -e $filename ) {
        my $t = $this->_readFile( $filename );
        my $lease = { split( /\r?\n/, $t ) };
        return $lease;
    }
    return undef;
}

# filenames for lock and lease files
sub _controlFileName {
    my( $this, $type ) = @_;

    my $fn = $this->{file};
    $fn =~ s/txt$/$type/;
    return $fn;
}

sub _saveFile {
    my( $this, $name, $text ) = @_;
    _mkPathTo( $name );
    open( FILE, '>'.$name ) ||
      throw Error::Simple( 'HG: failed to create .file '.$name.': '.$! );
    binmode( FILE ) ||
      throw Error::Simple( 'HG: failed to binmode '.$name.': '.$! );
    print FILE $text;
    close( FILE) ||
      throw Error::Simple( 'HG: failed to close file '.$name.': '.$! );

    return undef;
}

sub _mkPathTo {
    my( $file ) = @_;

    $file = TWiki::Sandbox::untaintUnchecked( $file ); 
    my $path = File::Basename::dirname($file);
    eval {
        File::Path::mkpath($path, 0, $TWiki::cfg{Hg}{dirPermission});
    };
    if ($@) {
       throw Error::Simple("HG: failed to create ${path}: $!");
    }
}

=pod

---++ ObjectMethod isLocked( ) -> ($user, $time)

See if a twiki lock exists. Return the lock user and lock time if it does.

=cut

sub isLocked {
    my( $this ) = @_;

    my $filename = $this->_controlFileName('lock');
    if ( -e $filename ) {
        my $t = $this->_readFile( $filename );
        return split( /\s+/, $t, 2 );
    }
    return ( undef, undef );
}

=pod

---++ ObjectMethod setLock($lock, $user)

Set a lock on the topic, if $lock, otherwise clear it.
$user is a wikiname.

SMELL: there is a tremendous amount of potential for race
conditions using this locking approach.

=cut

sub setLock {
    my( $this, $lock, $user ) = @_;

    $user = $this->{session}->{user} unless $user;

    my $filename = $this->_controlFileName('lock');
    if( $lock ) {
        my $lockTime = time();
        $this->_saveFile( $filename, $user."\n".$lockTime );
    } else {
        unlink $filename ||
          throw Error::Simple( 'HG: failed to delete '.$filename.': '.$! );
    }
}

sub addRevisionFromText {
    my( $this, $text, $comment, $user, $date ) = @_;
    $this->init();
    if( !$this->{version} ) {
	$this->{version} = $this->numRevisions() + 1;
    }
    unless( -e $this->{file} ) {
        $this->_ci( $comment, $user, $date );
    }
    $this->_saveFile( $this->{file}, $text );
    $this->_ci( $comment, $user, $date );
}

# Used in subclasses for late initialisation during object creation
# (after the object is blessed into the subclass)
sub init {
    my $this = shift;

    return unless $this->{topic};

    unless( -e $this->{file} ) {
	TWiki::Store::Mercurial::_mkPathTo( $this->{file} );
	# "touch" file to create it
	open FILE, ">$this->{file}";
	close FILE;
	#hg add command
	my ( $hgOutput, $exit ) =
	    $this->{session}->{sandbox}->sysCommand(
		    $TWiki::cfg{Hg}{initCmd}, FILENAME => $this->{file} );
	if( $exit ) {
	    throw Error::Simple( $TWiki::cfg{Hg}{initCmd}.
		    ' of '.$this->_hidePath($this->{file}).
		    ' failed: '.$hgOutput );
	} 
    }
}

sub _ci {
    my( $this, $comment, $user, $date ) = @_;

    $comment = "Version $this->{version}";#: $comment";
    my( $cmd, $hgOutput, $exit );
    chmod( $TWiki::cfg{Hg}{filePermission}, $this->{file} );
    if( defined( $date )) {
	($hgOutput, $exit)= $this->{session}->{sandbox}->sysCommand(
					       $TWiki::cfg{Hg}{ciDateCmd},
					       USERNAME => $user,
					       FILENAME => $this->{file},
					       COMMENT => $comment,
					       DATE => "$date -0000" );
    } else {
        $cmd = $TWiki::cfg{Hg}{ciCmd};
	($hgOutput, $exit)= $this->{session}->{sandbox}->sysCommand(
					       $TWiki::cfg{Hg}{ciCmd},
					       USERNAME => $user,
					       FILENAME => $this->{file},
					       COMMENT => $comment );
    }
    $hgOutput ||= '';
    
    if( $exit ) {
        throw Error::Simple($cmd.' of '.$this->_hidePath($this->{file}).
                              ' failed: '.$exit.' '.$hgOutput );
    }

}

# Chop out recognisable path components to prevent hacking based on error
# messages
sub _hidePath {
    my ( $this, $erf ) = @_;
    $erf =~ s#.*(/\w+/\w+\.[\w,]*)$#...$1#;
    return $erf;
}

=pod

---++ ObjectMethod readMetaData($name) -> $text
Get a meta-data block for this web

=cut

sub readMetaData {
    my( $this, $name ) = @_;
    my $file = $TWiki::cfg{DataDir}.'/'.$this->{web}.'/.'.$name;
    if( -e $file ) {
        return $this->_readFile( $file );
    }
    return '';
}

=pod

---++ ObjectMethod saveMetaData( $web, $name ) -> $text

Write a named meta-data string. If web is given the meta-data
is stored alongside a web.

=cut

sub saveMetaData {
    my ( $this, $name, $text ) = @_;

    my $file = $TWiki::cfg{DataDir}.'/'.$this->{web}.'/.'.$name;

    return $this->_saveFile( $file, $text );
}

=pod

---++ ObjectMethod searchInWebContent($searchString, $web, \@topics, \%options ) -> \%map

Search for a string in the content of a web. The search must be over all
content and all formatted meta-data, though the latter search type is
deprecated (use searchMetaData instead).

   * =$searchString= - the search string, in egrep format if regex
   * =$web= - The web to search in
   * =\@topics= - reference to a list of topics to search
   * =\%options= - reference to an options hash
The =\%options= hash may contain the following options:
   * =type= - if =regex= will perform a egrep-syntax RE search (default '')
   * =casesensitive= - false to ignore case (defaulkt true)
   * =files_without_match= - true to return files only (default false)

The return value is a reference to a hash which maps each matching topic
name to a list of the lines in that topic that matched the search,
as would be returned by 'grep'. If =files_without_match= is specified, it will
return on the first match in each topic (i.e. it will return only one
match per topic, and will not return matching lines).

=cut

sub searchInWebContent {
    my( $this, $searchString, $topics, $options ) = @_;
    ASSERT(defined $options) if DEBUG;
    my $type = $options->{type} || '';

    # I18N: 'grep' must use locales if needed,
    # for case-insensitive searching.  See TWiki::setupLocale.
    my $program = '';
    # FIXME: For Cygwin grep, do something about -E and -F switches
    # - best to strip off any switches after first space in
    # EgrepCmd etc and apply those as argument 1.
    if( $type eq 'regex' ) {
        $program = $TWiki::cfg{Hg}{EgrepCmd};
    } else {
        $program = $TWiki::cfg{Hg}{FgrepCmd};
    }

    $program =~ s/%CS{(.*?)\|(.*?)}%/$options->{casesensitive}?$1:$2/ge;
    $program =~ s/%DET{(.*?)\|(.*?)}%/$options->{files_without_match}?$2:$1/ge;

    my $sDir = $TWiki::cfg{DataDir}.'/'.$this->{web}.'/';
    my $seen = {};
    # process topics in sets, fix for Codev.ArgumentListIsTooLongForSearch
    my $maxTopicsInSet = 512; # max number of topics for a grep call
    my @take = @$topics;
    my @set = splice( @take, 0, $maxTopicsInSet );
    my $sandbox = $this->{session}->{sandbox};
    while( @set ) {
        @set = map { "$sDir/$_.txt" } @set;
        my ($matches, $exit ) = $sandbox->sysCommand(
            $program,
            TOKEN => $searchString,
            FILES => \@set);
        foreach my $match ( split( /\r?\n/, $matches )) {
            if( $match =~ m/([^\/]*)\.txt(:(.*))?$/ ) {
                push( @{$seen->{$1}}, $3 );
            }
        }
        @set = splice( @take, 0, $maxTopicsInSet );
    }
    return $seen;
}

sub addRevisionFromStream {
    my( $this, $stream, $comment, $user, $date ) = @_;
    $this->init();
    $this->_saveStream( $stream );
    $this->_ci( $comment, $user, $date );
}

sub _saveStream {
    my( $this, $fh ) = @_;

    ASSERT($fh) if DEBUG;

    _mkPathTo( $this->{file} );
    open( F, '>'.$this->{file} ) ||
        throw Error::Simple( 'HG: open '.$this->{file}.' failed: '.$! );
    binmode( F ) ||
      throw Error::Simple( 'HG: failed to binmode '.$this->{file}.': '.$! );
    my $text;
    binmode(F);
    while( read( $fh, $text, 1024 )) {
        print F $text;
    }
    close(F) ||
        throw Error::Simple( 'HG: close '.$this->{file}.' failed: '.$! );;

    chmod( $TWiki::cfg{Hg}{filePermission}, $this->{file} );

    return '';
}

sub replaceRevision {
# This can't be done, unless we store the version number in the comment. Which we do now ;-)
    my $this = shift;
    $this->{version} = $this->numRevisions();
    return $this->addRevisionFromText( @_ );
}

=pod

---++ ObjectMethod removeWeb( $web )
   * =$web= - web being removed

Destroy a web, utterly. Removed the data and attachments in the web.

Use with great care! No backup is taken!

=cut

sub removeWeb {
    my $this = shift;

    # Just make sure of the context
    ASSERT(!$this->{topic}) if DEBUG;

    _rmtree( $TWiki::cfg{DataDir}.'/'.$this->{web} );
    _rmtree( $TWiki::cfg{PubDir}.'/'.$this->{web} );
}

=pod

---++ ObjectMethod moveTopic( $newWeb, $newTopic )
Move/rename a topic.

=cut

sub moveTopic {
    my( $this, $newWeb, $newTopic ) = @_;

    my $oldWeb = $this->{web};
    my $oldTopic = $this->{topic};

    # Move data file
    my $new = new TWiki::Store::Mercurial( $this->{session},
                                         $newWeb, $newTopic, '' );
    # SMELL remove this once Hg can handle absolute paths for the dest
    # SMELL add data once all in hg
    $new->{file} =~ s/^$TWiki::cfg{DataDir}\///;

    _moveFile( $this->{file}, $new->{file}, $this );

    # Move attachments
    my $from = $TWiki::cfg{PubDir}.'/'.$this->{web}.'/'.$this->{topic};
    if( -e $from ) {
#SMELL add pub once all in hg
        my $to = $newWeb.'/'.$newTopic;
        _moveFile( $from, $to, $this );
    }
}

sub _moveFile {
    my( $from, $to, $this ) = @_;
    
    _mkPathTo( $to );
    
    my ($hgOut, $exit) =
      $this->{session}->{sandbox}->sysCommand(
	      $TWiki::cfg{Hg}{moveCmd},
	      SOURCE => $from,
	      DEST => $to );
   
    if( $exit ) {
        throw Error::Simple( 'HG: move '.$from.' to '.$to.' failed: '.$! );
    }
}

=pod

---++ ObjectMethod copyTopic( $newWeb, $newTopic )
Copy a topic.

=cut

sub copyTopic {
    my( $this, $newWeb, $newTopic ) = @_;

    my $oldWeb = $this->{web};
    my $oldTopic = $this->{topic};

    my $new = new TWiki::Store::Mercurial( $this->{session},
                                         $newWeb, $newTopic, '' );

    _copyFile( $this->{file}, $new->{file} );

    if( opendir(DIR, $TWiki::cfg{PubDir}.'/'.$this->{web}.'/'.
                  $this->{topic} )) {
        for my $att ( grep { !/^\./ } readdir DIR ) {
            $att = TWiki::Sandbox::untaintUnchecked( $att );
            my $oldAtt = new TWiki::Store::Mercurial(
                $this->{session}, $this->{web}, $this->{topic}, $att );
            $oldAtt->copyAttachment( $newWeb, $newTopic );
        }

        closedir DIR;
    }
}

sub _copyFile {
    my( $this, $from, $to ) = @_;

    _mkPathTo( $to );

    my ($hgOut, $exit) =
      $this->{session}->{sandbox}->sysCommand(
	      $TWiki::cfg{Hg}{copyCmd},
	      SOURCE => $from,
	      DEST => $to );
    
    if( $exit ) {
        throw Error::Simple( 'HG: copy '.$from.' to '.$to.' failed: '.$! );
    }
}

=pod

---++ ObjectMethod restoreLatestRevision($wikiname)

Restore the plaintext file from the revision at the head.
SMELL: This shortcuts the deleteRevision - restoreLatestRevision
SMELL: combo by going back to the last-1 revision.

=cut

sub deleteRevision { }

sub restoreLatestRevision {
    my( $this, $user ) = @_;

    my $rev = $this->numRevisions();
    if ( $rev > 1 ) {
	my $text = $this->getRevision( $rev-1 );

	$this->_saveFile( $this->{file}, $text );
    }
}

=pod

---++ ObjectMethod moveAttachment( $newWeb, $newTopic, $newAttachment )
Move an attachment from one topic to another. The name is retained.

=cut

sub moveAttachment {
    my( $this, $newWeb, $newTopic, $newAttachment ) = @_;

    # FIXME might want to delete old directories if empty
    my $new = TWiki::Store::Mercurial->new( $this->{session}, $newWeb,
                                          $newTopic, $newAttachment );

    _moveFile( $this->{file}, $new->{file} );

}

=pod

---++ ObjectMethod moveWeb(  $newWeb )
Move a web.

=cut

sub moveWeb {
    my( $this, $newWeb ) = @_;
    _moveFile( $TWiki::cfg{DataDir}.'/'.$this->{web},
               $TWiki::cfg{DataDir}.'/'.$newWeb );
    if( -d $TWiki::cfg{PubDir}.'/'.$this->{web} ) {
        _moveFile( $TWiki::cfg{PubDir}.'/'.$this->{web},
                   $TWiki::cfg{PubDir}.'/'.$newWeb );
    }
}

=pod

---+++ ObjectMethod getWorkArea( $key ) -> $directorypath

Gets a private directory uniquely identified by $key. The directory is
intended as a work area for plugins.

The standard is a directory named the same as "key" under
$TWiki::cfg{Hg}{WorkAreaDir}

SMELL should this be under Hg control as well? If not it can't be distributed

=cut

sub getWorkArea {
    my( $this, $key ) = @_;

    # untaint and detect nasties
    $key = TWiki::Sandbox::normalizeFileName( $key );
    throw Error::Simple( "Bad work area name $key" ) unless ( $key );

    my $dir =  "$TWiki::cfg{Hg}{WorkAreaDir}/$key";

    unless( -d $dir ) {
        mkdir( $dir ) || throw Error::Simple(
            'HG: failed to create '.$key.'work area: '.$! );
    }
    return $dir;
}

sub getRevisionAtTime {
    my( $this, $date ) = @_;

    my ($hgOutput, $exit) = $this->{session}->{sandbox}->sysCommand(
	    $TWiki::cfg{Hg}{rlogDateCmd},
	    DATE => "$date -0000",
	    FILENAME => $this->{file}
	    );

#SMELL fix this to use Version ci comments and a _getTWikiRevFromHgRev
    if ( $hgOutput ) {
	my $i=$this->numRevisions();
	for(; $i && $this->_getHgRev($i) != $hgOutput; $i--){}
        return $i+1;
    }
    return 1;
}

=pod

---++ ObjectMethod getStream() -> \*STREAM

Return a text stream that will supply the text stored in the topic.

=cut

sub getStream {
    my( $this ) = shift;
    my $strm;
    unless( open( $strm, '<'.$this->{file} )) {
        throw Error::Simple( 'HG: stream open '.$this->{file}.
                               ' failed: '.$! );
    }
    return $strm;
}

1;
